from ucimlrepo import fetch_ucirepo 
import pandas as pd
from sklearn.preprocessing import OneHotEncoder
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
import numpy as np
from sklearn.utils import resample
from sklearn.metrics import classification_report
from time import perf_counter
from tqdm import tqdm

def get_poker(train_size=0.8, test_size=0.2, verbose=False, seed=42, selected_labels=None):

    poker_hand = fetch_ucirepo(id=158) 
    
    df_x = poker_hand.data.features 
    df_y = poker_hand.data.targets['CLASS'].to_numpy(dtype=np.uint32)  

    s_cols = [col for col in df_x.columns if col.startswith('S')]
    c_cols = [col for col in df_x.columns if col.startswith('C')]

    s_encoder = OneHotEncoder(sparse_output=False)
    c_encoder = OneHotEncoder(sparse_output=False)

    s_binary = pd.DataFrame(
        s_encoder.fit_transform(df_x[s_cols]),
        columns=[f"{col}_{i}" for col in s_cols for i in range(1, 5)],
        index=df_x.index
    )

    c_binary = pd.DataFrame(
        c_encoder.fit_transform(df_x[c_cols]),
        columns=[f"{col}_{i}" for col in c_cols for i in range(1, 14)],
        index=df_x.index
    )

    binary_df_x = pd.concat([s_binary, c_binary], axis=1).to_numpy(dtype=np.uint8)


    if selected_labels is not None:
        mask = np.isin(df_y, selected_labels)
        binary_df_x = binary_df_x[mask]
        df_y = df_y[mask]
        
        label_map = {label: i for i, label in enumerate(sorted(np.unique(df_y)))}
        df_y = np.array([label_map[y] for y in df_y], dtype=np.uint32)

    x_train, x_test, y_train, y_test = train_test_split(binary_df_x, df_y, train_size=train_size, test_size=test_size, random_state=seed)

    if verbose:

        unique_values, counts = np.unique(y_test, return_counts=True)
        percentages = (counts / len(y_test)) * 100

        for value, percentage, count in zip(unique_values, percentages, counts):
            print(f"Class {value}: {percentage:.2f}% ({count})")

    return x_train, y_train, x_test, y_test


def get_match_statistics(large_set, small_set):
    
    match_count = 0
    for small_arr in tqdm(small_set, total=len(small_set)):
        if np.any(np.all(large_set == small_arr, axis=1)):
            match_count += 1
    
    percentage = (match_count / len(small_set)) * 100

    return {
        'total_matches': match_count,
        'percentage': percentage,
        'total_compared': len(small_set)
    }



def downsample_data(X, y, verbose=False):


    unique_classes, class_counts = np.unique(y, return_counts=True)
    
    min_count = np.min(class_counts)
    
    downsampled_X = []
    downsampled_y = []
    
    for cls in unique_classes:
        X_class = X[y == cls]
        y_class = y[y == cls]
        
        X_downsampled, y_downsampled = resample(
            X_class, 
            y_class, 
            replace=False,  # Sample without replacement
            n_samples=min_count, 
            random_state=42  # For reproducibility
        )
        
        downsampled_X.append(X_downsampled)
        downsampled_y.append(y_downsampled)
    
    X_balanced = np.vstack(downsampled_X, dtype=np.uint8)
    y_balanced = np.concatenate(downsampled_y, dtype=np.uint32)
    
    if verbose:

        unique_values, counts = np.unique(y_balanced, return_counts=True)
        percentages = (counts / len(y_balanced)) * 100

        for value, percentage, count in zip(unique_values, percentages, counts):
            print(f"Class {value}: {percentage:.2f}% ({count})")


    return X_balanced, y_balanced



if __name__ == "__main__":

    st = perf_counter()
    x_train, y_train, x_test, y_test = get_poker(train_size=10000, selected_labels=[0, 1])
    et = perf_counter()
    print(et - st)
    st = perf_counter()
    x_train, y_train = downsample_data(x_train, y_train, verbose=False)
    et = perf_counter()
    print(et - st)

    st = perf_counter()
    stats = get_match_statistics(x_train, x_test)
    et = perf_counter()
    print(et - st)
    

    print(stats)

    # clf = RandomForestClassifier()
    # clf.fit(x_train, y_train)

    # pred = clf.predict(x_test)

    # print(np.mean(pred == y_test))


    # print(classification_report(y_test, pred))


from numba import njit, prange
import numpy as np
from pytsetlin.core.feedback import evaluate_clauses_training, get_update_p, update_clauses, evaluate_clause
from pytsetlin.core import config


@njit
def train_epoch(cb, wb, x, y, threshold, s_min_inv, s_inv, n_outputs, n_literals, n_literal_budget, boost_true_positives):
    

    cb = np.ascontiguousarray(cb)
    wb = np.ascontiguousarray(wb)

    indices = np.arange(x.shape[0], dtype=np.int32) # this does not need to be calulated at every epoch?
    np.random.shuffle(indices)

    clause_outputs = np.ones(cb.shape[0], dtype=np.uint8)
    literals_counts = np.zeros(cb.shape[0], dtype=np.uint32)

    y_hat = np.zeros(y.shape, dtype=np.uint32)

    for indice in indices:
        literals = x[indice]
        target = y[indice]
        
        evaluate_clauses_training(literals, cb, n_literals, clause_outputs, literals_counts)
        
        vote_values = np.dot(wb.astype(np.float32), clause_outputs.astype(np.float32))

        not_target = np.random.randint(0, n_outputs)

        while(not_target == target):
            not_target = np.random.randint(0, n_outputs)

        votes = vote_values[target]

        v_clamped_pos = np.clip(np.array(votes), -threshold, threshold)

        pos_update_p =  (threshold - v_clamped_pos) / (2*threshold)

        votes = vote_values[not_target]

        v_clamped_neg = np.clip(np.array(votes), -threshold, threshold)

        neg_update_p =  (threshold + v_clamped_neg) / (2*threshold)

        update_clauses(cb, wb, clause_outputs, literals_counts, pos_update_p, neg_update_p, target, 
                    not_target, literals, n_literals, n_literal_budget, s_min_inv, s_inv, boost_true_positives)

        y_hat[indice] = np.argmax(vote_values)

    return y_hat


@njit
def classify(x, clause_block, weight_block, n_literals):

    clause_outputs = evaluate_clause(x, clause_block, n_literals)

    class_sums = np.dot(weight_block.astype(np.float32), clause_outputs.astype(np.float32))

    return np.argmax(class_sums)
  
@njit(parallel=config.OPERATE_PARALLEL)
def eval_predict(x, cb, wb, n_literals):
    y_pred = np.zeros(x.shape[0], dtype=np.uint32)
    
    for i in prange(x.shape[0]):
        y_pred[i] = classify(x[i], cb, wb, n_literals)
    
    return y_pred



import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from pytsetlin.data.parity import get_parity
from pytsetlin import TsetlinMachine






if __name__ == "__main__":

    x, y = get_parity(n_rows = 1000)
    xt, yt = get_parity(n_rows = 200)


    tm = TsetlinMachine(n_clauses=100,
                        threshold=200,
                        s=2.0)

    tm.set_train_data(x, y)

    tm.set_eval_data(xt, yt)

    tm.train(training_epochs=100)





import numpy as np


def get_parity(n_rows=1000, n_digits=10):


    x = np.random.choice(a = [0, 1], size=(n_rows, n_digits)).astype(np.uint8)

    y = np.sum(x, axis=1, dtype=np.uint32) % 2

    return x, y


if __name__ == "__main__":

    get_parity(n_digits=3, n_rows=2)



import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import pytest
import numpy as np
from pytsetlin.tsetlin_machine import TsetlinMachine
from pytsetlin.data import xor

@pytest.fixture
def sample_binary_data():

    X_train, y_train = xor.get_xor(n_rows = 1000, noise_fraction=0.3)
    X_test, y_test = xor.get_xor(n_rows = 200, noise_fraction=0.0)

    return X_train, y_train, X_test, y_test

@pytest.fixture
def basic_tsetlin():

    return TsetlinMachine(n_clauses=100, s=2.0, threshold=200)

def test_initialization(basic_tsetlin):

    assert basic_tsetlin.n_clauses == 100
    assert basic_tsetlin.s == 2.0
    assert basic_tsetlin.threshold == 200
    assert basic_tsetlin.C is None
    assert basic_tsetlin.W is None

def test_data_validation(basic_tsetlin):

    invalid_x = np.random.rand(100, 10)  
    invalid_y = np.random.randint(2, size=50)  
    valid_x = np.random.randint(2, size=(100, 10), dtype=np.uint8)
    valid_y = np.random.randint(2, size=100, dtype=np.uint32)
    
    with pytest.raises(ValueError):
        basic_tsetlin.set_train_data(invalid_x, valid_y)
    
    with pytest.raises(ValueError):
        basic_tsetlin.set_train_data(valid_x, invalid_y)

def test_memory_allocation(basic_tsetlin, sample_binary_data):

    X_train, y_train, _, _ = sample_binary_data
        
    basic_tsetlin.set_train_data(X_train, y_train)
    basic_tsetlin.allocate_memory()
    
    expected_literal_count = X_train.shape[1] * 2  
    assert basic_tsetlin.C.shape == (basic_tsetlin.n_clauses, expected_literal_count)
    assert basic_tsetlin.W.shape == (basic_tsetlin.n_outputs, basic_tsetlin.n_clauses)


def test_training_performance(basic_tsetlin, sample_binary_data):

    X_train, y_train, X_test, y_test = sample_binary_data
    
    basic_tsetlin.set_train_data(X_train, y_train)
    basic_tsetlin.set_eval_data(X_test, y_test)
    
    results = basic_tsetlin.train(
        training_epochs=100,
        eval_freq=1,
        hide_progress_bar=True
    )
    
    assert 'train_time' in results
    assert 'eval_acc' in results
    

    predictions = np.zeros(X_test.shape[0])

    for i in range(X_test.shape[0]):

        predictions[i] = basic_tsetlin.predict(X_test[i])

    assert predictions.shape == y_test.shape
    assert np.all((predictions == 0) | (predictions == 1))  # binary predictions, this i can add as check to the set_ funcs
    
    accuracy = np.mean(predictions == y_test)
    assert accuracy == 1.0  



def test_singal_eval(basic_tsetlin):

    C = np.array([[100, -100], [-100, 100]])

    x = np.array([1])

    clause_outputs = basic_tsetlin.evaluate_clauses(x, memory=C)

    assert clause_outputs[0] == 1 
    assert clause_outputs[1] == 0

def test_custom_save_best_filename(basic_tsetlin, sample_binary_data):
    X_train, y_train, X_test, y_test = sample_binary_data
    
    basic_tsetlin.set_train_data(X_train, y_train)
    basic_tsetlin.set_eval_data(X_test, y_test)
    
    tmp_path = "saved_states"
    custom_filename = "my_custom_best_model.npz"
    custom_path = os.path.join(tmp_path, custom_filename)
    
    assert not os.path.exists(custom_path)
    
    results = basic_tsetlin.train(
        training_epochs=10,
        eval_freq=1,
        hide_progress_bar=True,
        save_best_state=True,
        file_name = custom_filename,
        location_dir = tmp_path
    )
    
    assert os.path.exists(custom_path)
    
    # Second training run - should raise FileExistsError
    with pytest.raises(FileExistsError):
        basic_tsetlin.train(
            training_epochs=10,
            eval_freq=1,
            hide_progress_bar=True,
            save_best_state=True,
            file_name= custom_filename,
            location_dir=tmp_path
        )

    if os.path.exists(custom_path):
        os.remove(custom_path)



from numba import njit, prange
import numpy as np
from pytsetlin.core import config


@njit(parallel=config.OPERATE_PARALLEL)
def evaluate_clauses_training(literals, cb, n_literals, clause_outputs, literals_counts):

    # captures the imply operation ta -> lit?

    clause_outputs.fill(1)

    literals_counts.fill(0)
    
    for clause_k in prange(cb.shape[0]):

        pos_literal_count = 0
        neg_literal_count = 0

        for literal_k in range(n_literals):

            if(cb[clause_k, literal_k] > 0):


                if(literals[literal_k] == 0):

                    clause_outputs[clause_k] = 0

                    break
            
                pos_literal_count += 1

            if(cb[clause_k, literal_k + n_literals] > 0):


                if(literals[literal_k] == 1):   

                    clause_outputs[clause_k] = 0

                    break

                neg_literal_count += 1

        literals_counts[clause_k] = pos_literal_count + neg_literal_count



@njit
def evaluate_clause(literals, clause_block, n_literals):

    clause_outputs = np.ones(clause_block.shape[0], dtype=np.uint8)

    for clause_k in range(clause_block.shape[0]):

        is_empty_clause = True

        for literal_k in range(n_literals):

            if(clause_block[clause_k, literal_k] > 0):

                is_empty_clause = False

                if(literals[literal_k] == 0):

                    clause_outputs[clause_k] = 0

                    break
            
            if(clause_block[clause_k, literal_k + n_literals] > 0):

                is_empty_clause = False

                if(literals[literal_k] == 1):    

                    clause_outputs[clause_k] = 0

                    break

        if(is_empty_clause):

            clause_outputs[clause_k] = 0


    return clause_outputs


@njit
def get_update_p(vote_values, threshold, y, target_class):
  
    vote_value = np.clip(np.array(vote_values[y]), -threshold, threshold)

    if target_class:
        return (threshold - vote_value) / (2 * threshold)
    else:
        return (threshold + vote_value) / (2 * threshold)



@njit(parallel=config.OPERATE_PARALLEL)
def update_clauses(cb, wb, clause_outputs, literals_counts, positive_prob, negative_prob, target, not_target, literals, n_literals, n_literal_budget, s_min_inv, s_inv, boost_true_positives):
    
    for clause_k in prange(cb.shape[0]):

        if literals_counts[clause_k] > n_literal_budget:
            clause_outputs[clause_k] = 0

        if np.random.random() <= positive_prob:
            update_clause(cb, wb, 1, literals, n_literals, clause_outputs, clause_k, target, s_min_inv, s_inv, boost_true_positives) 

        if np.random.random() <= negative_prob:
            update_clause(cb, wb, -1, literals, n_literals, clause_outputs, clause_k, not_target, s_min_inv, s_inv, boost_true_positives) 
 



@njit
def update_clause(cb, wb, target, literals, n_literals, clause_output, clause_k, class_k, s_min_inv, s_inv, boost_true_positives):
    
    sign = 1 if (wb[class_k, clause_k] >= 0) else -1

    if(target*sign > 0):

        if clause_output[clause_k] == 1:

            wb[class_k, clause_k] += sign

            T1aFeedback(cb, clause_k, literals, n_literals, s_min_inv, s_inv, boost_true_positives)

        else:
            
            T1bFeedback(cb, clause_k, n_literals, s_inv)

    elif(target*sign < 0):

        if clause_output[clause_k] == 1:

            wb[class_k, clause_k] -= sign

            T2Feedback(cb, clause_k, literals, n_literals)


@njit
def T1aFeedback(cb, clause_k, literals, n_literals, s_min_inv, s_inv, boost_true_positives):
    
    upper_state =  127
    lower_state = -127

    for literal_k in range(n_literals):

        if(literals[literal_k] == 1):

            # here we boost true possitives
            if boost_true_positives:
                if(cb[clause_k, literal_k] < upper_state):
                    cb[clause_k, literal_k] += 1

            else:
                if(np.random.random() <= s_min_inv):
                    if(cb[clause_k, literal_k] < upper_state):
                        cb[clause_k, literal_k] += 1

            if(np.random.random() <= s_inv):
                if(cb[clause_k, literal_k + n_literals] > lower_state):
                    cb[clause_k, literal_k + n_literals] -= 1          
            

        else:
            if(np.random.random() <= s_inv):
                if(cb[clause_k, literal_k] > lower_state):
                    cb[clause_k, literal_k] -= 1

            # here we boost true possitives
            if boost_true_positives:
                if(cb[clause_k, literal_k + n_literals] < upper_state):
                    cb[clause_k, literal_k + n_literals] += 1       

            else:
                if(np.random.random() <= s_min_inv):
                    if(cb[clause_k, literal_k + n_literals] < upper_state):
                        cb[clause_k, literal_k + n_literals] += 1                           

@njit
def T1bFeedback(cb, clause_k, n_literals, s_inv):

    lower_state = -127
    
    for literal_k in range(n_literals):

        if np.random.random() <= s_inv:

            if cb[clause_k, literal_k] > lower_state:
                cb[clause_k, literal_k] -= 1

        if np.random.random() <= s_inv:

            if cb[clause_k, literal_k + n_literals] > lower_state:
                cb[clause_k, literal_k + n_literals] -= 1

@njit
def T2Feedback(cb, clause_k, literals, n_literals):

    for literal_k in range(n_literals):

        if(literals[literal_k] == 0):
            if(cb[clause_k, literal_k] <= 0):
                cb[clause_k, literal_k] += 1

        else:
            if(cb[clause_k, literal_k + n_literals] <= 0):
                
                cb[clause_k, literal_k + n_literals] += 1


from pytsetlin.core import executor
from pytsetlin.core import config
from pytsetlin.core import tuner

from numba import set_num_threads
import tqdm
import numpy as np
import optuna

from time import perf_counter
import os

import logging
logging.basicConfig(level=logging.INFO, format='%(levelname)s - %(message)s')




class TsetlinMachine:
    def __init__(self,
                 n_clauses:int = 50,
                 s:float = 5.0,
                 threshold:int = 100,
                 n_literal_budget:int = np.inf,
                 boost_true_positives:bool = True, 
                 n_threads:int = 1):

        self.n_clauses = n_clauses        
        self.s = s    
        self.threshold = threshold    
        self.n_literal_budget = n_literal_budget
        self.boost_true_positives = boost_true_positives    

        self.C = None        
        self.W = None

        self.x_train = None
        self.y_train = None

        self.x_eval = None
        self.y_eval = None

        config.N_THREADS = n_threads
        config.OPERATE_PARALLEL = True if config.N_THREADS > 1 and not config.N_THREADS <= 0 else False
        set_num_threads(config.N_THREADS)


    def set_train_data(self, instances:np.array, targets:np.array):
        

        if not isinstance(instances, np.ndarray):
            raise ValueError("x_train must be of type np.ndarray, x_train type: {}".format(type(instances)))

        if instances.shape[0] != targets.shape[0]:
            raise ValueError("Data x_train and y_train must have the same number of examples: {} != {}".format(instances.shape[0], targets.shape[0]))

        if instances.dtype != np.uint8:
            raise ValueError("Data x_train must be of type np.uint8, was: {}".format(instances.dtype))

        if targets.dtype != np.uint32:
            raise ValueError("Data y_train must be of type np.uint32, was: {}".format(targets.dtype))
        

        self.n_literals = instances.shape[1]
        self.n_outputs = len(np.unique(targets)) 
        
        self.x_train = instances
        self.y_train = targets



    def set_eval_data(self, instances:np.array, targets:np.array):
        
        if not isinstance(instances, np.ndarray):
            raise ValueError("x_eval must be of type np.ndarray, x_eval type: {}".format(type(instances)))

        if instances.shape[0] != targets.shape[0]:
            raise ValueError("Data x_eval and y_eval must have the same number of examples: {} != {}".format(instances.shape[0], targets.shape[0]))

        if instances.dtype != np.uint8:
            raise ValueError("Data x_eval must be of type np.uint8, was: {}".format(instances.dtype))

        if targets.dtype != np.uint32:
            raise ValueError("Data y_eval must be of type np.uint32, was: {}".format(targets.dtype))
        

        self.x_eval = instances
        self.y_eval = targets

    
    def allocate_memory(self):


        if (self.n_literals is None) or (self.n_outputs is None):
            raise ValueError("failed to allocate memory, make sure data is set using set_train_data() and set_eval_data()")


        self.C = np.zeros((self.n_clauses, 2*self.n_literals), dtype=np.int8)
        self.W = np.random.choice(np.array([-1, 1]), size=(self.n_outputs, self.n_clauses), replace=True).astype(np.int32)


        if self.s < 1.0:
            raise ValueError('s must be larger or equal to 1.0. s >= 1.0')


        self.s_min_inv = (self.s - 1.0) / self.s
        self.s_inv = (1.0 / self.s)




    def reset(self):
        self.C = None        
        self.W = None

        self.x_train = None
        self.y_train = None

        self.x_eval = None
        self.y_eval = None


    def train(self, 
              training_epochs:int=10,
              eval_freq:int=1,
              hide_progress_bar:bool=False,
              early_stop_at:float=100.0,
              save_best_state=False,
              file_name="tm_state.npz",
              location_dir="saved_states"):


        full_path = os.path.join(location_dir, file_name)
        if os.path.exists(full_path):
            raise FileExistsError(f"File already exists: {full_path}")


        self.allocate_memory()

        r = {
            'train_time' : [],
            'train_acc' : [],
            'eval_acc' : [],
        }

        train_score = "N/A"
        eval_score = "N/A"
        best_eval_acc = "N/A"
        best_eval_epoch = "#"
        
        with tqdm.tqdm(total=training_epochs, disable=hide_progress_bar) as progress_bar:
            progress_bar.set_description(f"Train Acc: {train_score}, Eval Acc: {eval_score}, Best Eval Acc: {best_eval_acc} ({best_eval_epoch})")


            for epoch in range(training_epochs):

                st = perf_counter() 

                train_predictions = executor.train_epoch(
                                        self.C, self.W, self.x_train, self.y_train, 
                                        self.threshold, self.s_min_inv, self.s_inv, self.n_outputs, self.n_literals, self.n_literal_budget,
                                        self.boost_true_positives)
                                    
                et = perf_counter()
                
                if (epoch+1) % eval_freq == 0:

                    train_score = round(100 * np.mean(train_predictions == self.y_train).item(), 2)
                    r["train_acc"].append(train_score)

                    eval_predictions = executor.eval_predict(self.x_eval, self.C, self.W, self.n_literals)
                    eval_score = round(100 * np.mean(eval_predictions == self.y_eval).item(), 2)
                    r["eval_acc"].append(eval_score)

                    if best_eval_acc == 'N/A' or eval_score > best_eval_acc:
                        best_eval_acc = eval_score
                        best_eval_epoch = epoch+1
                        r["best_eval_acc"] = best_eval_acc
                        r["best_eval_epoch"] = best_eval_epoch


                        if save_best_state:
                            self.save_state(file_name=file_name, location_dir=location_dir)

                r["train_time"].append(round(et-st, 2))

                progress_bar.set_description(f"Train Acc: {train_score}, Eval Acc: {eval_score}, Best Eval Acc: {best_eval_acc} ({best_eval_epoch})") 
                progress_bar.update(1)

                if not eval_score == 'N/A':
                    if eval_score >= early_stop_at:
                        break

        return r
    

    def predict(self, x):
        
        return executor.classify(x, self.C, self.W, self.n_literals)


    def evaluate_clauses(self, literals, memory:np.array=None):

        if memory is None and self.C is None:
            raise ValueError('did not find loaded input memory or trained memory. either input a loaded memory or train tm.')

        if self.C is not None:
            memory = self.C

        return executor.evaluate_clause(literals, memory, literals.shape[0])



    def save_state(self, file_name = 'tm_state.npz', location_dir='saved_states'):

        if not os.path.isdir(f'{location_dir}'):
            os.mkdir(f'{location_dir}')

        np.savez(f"{location_dir}/{file_name}", C=self.C, W=self.W)


    def optimize_hyperparameters(self, n_trials=100, training_epochs=10):

        logging.info(f"This optimization only tunes threshold, s, boost_true_positives, and n_literal_budget. n_clauses is fixed at {self.n_clauses}.")

        self.allocate_memory()

        study = optuna.create_study(direction='maximize')

        study.optimize(lambda trial: tuner.objective(trial, self, training_epochs), n_trials=n_trials, show_progress_bar=True)

        best_params = study.best_params.copy()

        best_value = study.best_value
        
        if 'do_literal_budget' in best_params:
            if not best_params['do_literal_budget']:
                best_params['n_literal_budget'] = "infinity"
            del best_params['do_literal_budget']
        
        print(f"Score: {best_value:.4f}")
        print("Parameters:")
        for param, value in best_params.items():
            print(f"  {param}: {value}")






if __name__ == "__main__":

    tm = TsetlinMachine(n_clauses=500)

    tm.optimize_hyperparameters()



    
from sklearn.datasets import fetch_openml
import numpy as np



def get_mnist():


    X, y = fetch_openml('mnist_784', version=1, return_X_y=True, as_frame=False)

    X = np.where(X.reshape((X.shape[0], 28 * 28)) > 75, 1, 0).astype(np.uint8)

    y = y.astype(np.uint32)

    x_train, x_test = X[:60000], X[60000:]
    y_train, y_test = y[:60000], y[60000:]


    return x_train, x_test, y_train, y_test



if __name__ == "__main__":

    X_train, X_test, y_train, y_test = get_mnist()




import numpy as np


def get_xor(n_rows=1000, 
            number_features=6, 
            noise_fraction=0.3, 
            seed=42):
    
    rng = np.random.default_rng(seed)
    X = rng.integers(0, 2, size=(n_rows, number_features))
    y = np.logical_xor(X[:, 0:1], X[:, 1:2])
    flips = rng.random((n_rows, 1)) < noise_fraction
    y = np.logical_xor(y, flips)
   
    y = y*1

    return X.astype(np.uint8), y.reshape(y.shape[0]).astype(np.uint32)


if __name__ == "__main__":

    x, y = get_xor()




import datasets
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_selection import chi2
from sklearn.feature_selection import SelectKBest
import numpy as np
from sklearn.utils import shuffle



def get_imdb(train_size=None, test_size=None, seed=42, max_features=20000, features=1000, max_df=0.80, min_df=3, ngram_range_max=3):

    imdb = datasets.load_dataset('imdb')
    x_train, y_train, x_test, y_test = imdb['train']['text'], imdb['train']['label'], imdb['test']['text'], imdb['test']['label']


    if train_size:
        x_train, y_train = shuffle(x_train, y_train, n_samples=train_size, random_state=seed)
    if test_size:
        x_test, y_test = shuffle(x_test, y_test, n_samples=test_size, random_state=seed)

    vectorizer = CountVectorizer(
        analyzer = 'word',
        binary=True,
        ngram_range=(1, ngram_range_max),
        max_features=max_features,
        max_df=max_df,
        min_df=min_df,
    )


    x_train = vectorizer.fit_transform(x_train).toarray().astype(np.uint8)
    y_train = np.array(y_train).astype(np.uint32)
    x_test = vectorizer.transform(x_test).toarray().astype(np.uint8)
    y_test = np.array(y_test).astype(np.uint32)

    
    SKB = SelectKBest(score_func=chi2, k=features)
    SKB.fit(x_train, y_train)

    x_train = SKB.transform(x_train)
    x_test = SKB.transform(x_test)

    return x_train, y_train, x_test, y_test


import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from pytsetlin.data.imdb import get_imdb
from pytsetlin import TsetlinMachine




if __name__ == "__main__":

    x_train, y_train, x_test, y_test = get_imdb()

    tm = TsetlinMachine(n_clauses=100,
                        threshold=500,
                        s=10.0,
			n_literal_budget=5)

    tm.set_train_data(x_train, y_train)

    tm.set_eval_data(x_test, y_test)

    r = tm.train(training_epochs=100)
    
    print(r)





import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from pytsetlin.data.mnist import get_mnist
from pytsetlin import TsetlinMachine




if __name__ == "__main__":

    X_train, X_test, y_train, y_test = get_mnist()

    tm = TsetlinMachine(n_clauses=500,
                        threshold=625,
                        s=10.0)

    tm.set_train_data(X_train, y_train)

    tm.set_eval_data(X_test, y_test)

    r = tm.train()
    
    print(r)





import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from pytsetlin.data.xor import get_xor
from pytsetlin import TsetlinMachine



import numpy as np


if __name__ == "__main__":

    # x, y = get_xor(n_rows = 1000, noise_fraction=0.3)
    # xt, yt = get_xor(n_rows = 200, noise_fraction=0.0)

    # xor gate
    x = np.array([[0, 0],
                  [0, 1],
                  [1, 0],
                  [1, 1]])

    y = np.array([0, 1, 1, 0])

    tm = TsetlinMachine(n_clauses=4)

    tm.set_train_data(x, y)
    tm.set_eval_data(x, y)

    tm.train()

    print(tm.C)
    print(tm.W)





import numpy as np
import optuna

def objective(trial, tm, training_epochs):

    tm.allocate_memory()

    tm.threshold = trial.suggest_int('threshold', 500, 20000, step=500)

    tm.s = trial.suggest_float('s', 1.0, 30.0, step=0.5)

    tm.boost_true_positives = trial.suggest_categorical('boost_true_positives', [True, False])
    
    do_literal_budget = trial.suggest_categorical('do_literal_budget', [True, False])


    if do_literal_budget:

        tm.n_literal_budget = trial.suggest_int('n_literal_budget', 10, 100, step=10)

    else:

        tm.n_literal_budget = np.inf    


    try:

        results = tm.train(hide_progress_bar=True, training_epochs=training_epochs)

    except:

        trial.report(0.0, step=0) 
        raise optuna.TrialPruned()


    return results['best_eval_acc']



import sys
import os
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from pytsetlin.data.poker import get_poker, downsample_data
from pytsetlin import TsetlinMachine




if __name__ == "__main__":

    x_train, y_train, x_test, y_test = get_poker(selected_labels=[0, 1, 2, 3])
    x_train, y_train = downsample_data(x_train, y_train, verbose=True)

    tm = TsetlinMachine(n_clauses=1000,
                        threshold=9650,
                        s=2.5, 
                        n_threads=6)
                                
    tm.set_train_data(x_train, y_train)

    tm.set_eval_data(x_test, y_test)

    r = tm.train(training_epochs=1000)
    
    print(r)


